import tensorflow as tf

class RiskConcentrationFunction:
    def __init__(self, portfolio):
        self.portfolio = portfolio

    def evaluate(self):
        return tf.reduce_sum(tf.square(self.risk_concentration_vector()))

    # the vector g in Feng & Palomar 2015
    def risk_concentration_vector(self):
        raise NotImplementedError("this method should be implemented in the child class")

    # jacobian of the vector function risk_concentration_vector with respect to weights
    def jacobian_risk_concentration_vector(self):
        raise NotImplementedError("this method should be implemented in the child class")
        # I'm not gonna use autograd here. We have the derivatives analytically already.
        #with tf.GradientTape() as t:
        #    t.watch(self.portfolio.weights)
        #    risk_vec = self.risk_concentration_vector()
        #return t.jacobian(risk_vec, self.portfolio.weights)


class RiskContribOverBudgetDoubleIndex(RiskConcentrationFunction):
    def risk_concentration_vector(self):
        N = len(self.portfolio.weights)
        marginal_risk = tf.math.multiply(self.portfolio.weights,
                tf.linalg.matvec(self.portfolio.covariance, self.portfolio.weights))
        normalized_marginal_risk = tf.math.divide(marginal_risk, self.portfolio.budget)
        return tf.tile(normalized_marginal_risk, [N]) - repeat(normalized_marginal_risk, N)


class RiskContribOverVarianceMinusBudget(RiskConcentrationFunction):
    def risk_concentration_vector(self):
        marginal_risk = tf.math.multiply(self.portfolio.weights,
                tf.linalg.matvec(self.portfolio.covariance, self.portfolio.weights))
        return marginal_risk / tf.reduce_sum(marginal_risk) - self.portfolio.budget

    def jacobian_risk_concentration_vector(self):
        Sigma_w = tf.linalg.matvec(self.portfolio.covariance, self.portfolio.weights)
        r = tf.multiply(self.portfolio.weights, Sigma_w)
        sum_r = tf.reduce_sum(r)
        Ut = tf.linalg.diag(Sigma_w) + tf.multiply(self.portfolio.covariance, self.portfolio.weights)
        return Ut / sum_r - 2 / (sum_r ** 2) * (r[..., None] * Sigma_w[..., None])

def repeat(vector, times):
    vector = tf.reshape(tf.tile(tf.reshape(vector, [-1, 1]), [1, times]), [-1])
    return vector




import tensorflow as tf
import numpy as np
from tqdm import tqdm
try:
    import quadprog
except ImportError:
    import warnings
    warnings.warn("not able to import quadprog."
                  " the successive convex optimizer wont work.")

__all__ = ['SuccessiveConvexOptimizer']


class SuccessiveConvexOptimizerValidator:

    def __init__(self):
        pass

    @property
    def gamma(self):
        return self._gamma

    @gamma.setter
    def gamma(self, value):
        if not (0 < value < 1):
            raise ValueError("gamma has to be a real number in the interval"
                             "(0, 1), got {}.".format(value))
        else:
            try:
                self._gamma = float(value)
            except Exception as e:
                raise e

    @property
    def zeta(self):
        return self._zeta

    @zeta.setter
    def zeta(self, value):
        if value < 0:
            raise ValueError("zeta has to be a positive real number, got {}.".format(value))
        else:
            try:
                self._zeta = float(value)
            except Exception as e:
                raise e

    @property
    def funtol(self):
        return self._funtol

    @funtol.setter
    def funtol(self, value):
        if value < 0:
            raise ValueError("funtol has to be a positive real number, got {}.".format(value))
        else:
            try:
                self._funtol = float(value)
            except Exception as e:
                raise e

    @property
    def wtol(self):
        return self._wtol

    @wtol.setter
    def wtol(self, value):
        if value < 0:
            raise ValueError("wtol has to be a positive real number, got {}.".format(value))
        else:
            try:
                self._wtol = float(value)
            except Exception as e:
                raise e
    @property
    def maxiter(self):
        return self._maxiter

    @maxiter.setter
    def maxiter(self, value):
        if value < 0:
            raise ValueError("maxiter has to be a positive integer, got {}.".format(value))
        else:
            try:
                self._maxiter = int(value)
            except Exception as e:
                raise e


class SuccessiveConvexOptimizer:
    """
    Successive Convex Approximation optimizer tailored for the risk parity problem.
    """
    def __init__(self, portfolio, tau = None, gamma = 0.9, zeta = 1E-7,
                 funtol = 1E-6, wtol = 1E-6, maxiter = 500, Cmat = None,
                 cvec = None, Dmat = None, dvec = None):
        self.portfolio = portfolio
        self.tau       = (tau or 0.05 * np.sum(np.diag(self.portfolio.covariance.numpy()))
                                 / (2 * self.portfolio.number_of_assets))
        sca_validator  = SuccessiveConvexOptimizerValidator()
        self.gamma     = sca_validator.gamma     = gamma
        self.zeta      = sca_validator.zeta      = zeta
        self.funtol    = sca_validator.funtol    = funtol
        self.wtol      = sca_validator.wtol      = wtol
        self.maxiter   = sca_validator.maxiter   = maxiter
        self.Cmat      = Cmat
        self.Dmat      = Dmat
        self.cvec      = cvec
        self.dvec      = dvec
        self.CCmat     = np.vstack((self.Cmat, self.Dmat)).T
        self.bvec      = np.concatenate((self.cvec, self.dvec))
        self.meq       = self.Cmat.shape[0]
        self._funk = self.get_objective_function_value()
        self.objective_function = [self._funk.numpy()]
        self._tauI = self.tau * tf.eye(self.portfolio.number_of_assets, dtype=tf.float64)

    @property
    def Cmat(self):
        return self._Cmat

    @Cmat.setter
    def Cmat(self, value):
        if value is None:
            self._Cmat = np.atleast_2d(np.ones(self.portfolio.number_of_assets))
        elif np.atleast_2d(value).shape[1] == self.portfolio.number_of_assets:
            self._Cmat = np.atleast_2d(value)
        else:
            raise ValueError("Cmat shape {} doesnt agree with the number of"
                             "assets {}".format(value.shape, self.number_of_assets))

    @property
    def Dmat(self):
        return self._Dmat

    @Dmat.setter
    def Dmat(self, value):
        if value is None:
            self._Dmat = np.eye(self.portfolio.number_of_assets)
        elif np.atleast_2d(value).shape[1] == self.portfolio.number_of_assets:
            self._Dmat = -np.atleast_2d(value)
        else:
            raise ValueError("Dmat shape {} doesnt agree with the number of"
                             "assets {}".format(value.shape, self.number_of_assets))

    @property
    def cvec(self):
        return self._cvec

    @cvec.setter
    def cvec(self, value):
        if value is None:
            self._cvec = np.array([1])
        elif len(value) == self.Cmat.shape[0]:
            self._cvec = value
        else:
            raise ValueError("cvec shape {} doesnt agree with Cmat shape"
                             "{}".format(value.shape, self.Cmat.shape))

    @property
    def dvec(self):
        return self._dvec

    @dvec.setter
    def dvec(self, value):
        if value is None:
            self._dvec = np.zeros(self.portfolio.number_of_assets)
        elif len(value) == self.Dmat.shape[0]:
            self._dvec = -value
        else:
            raise ValueError("dvec shape {} doesnt agree with Dmat shape"
                             "{}".format(value.shape, self.Dmat.shape))

    def get_objective_function_value(self):
        obj = self.portfolio.risk_concentration.evaluate()
        if self.portfolio.has_mean_return:
            obj -= self.portfolio.alpha * self.portfolio.mean_return
        if self.portfolio.has_variance:
            obj += self.portfolio.lmd * self.portfolio.variance
        return obj

    def iterate(self):
        wk = self.portfolio.weights
        g = self.portfolio.risk_concentration.risk_concentration_vector()
        A = self.portfolio.risk_concentration.jacobian_risk_concentration_vector()
        At = tf.transpose(A)
        Q = 2 * At @ A + self._tauI
        q = 2 * tf.linalg.matvec(At, g) - tf.linalg.matvec(Q, wk)
        if self.portfolio.has_variance:
            Q += self.portfolio.lmd * self.portfolio.covariance
        if self.portfolio.has_mean_return:
            q -= self.portfolio.alpha * self.portfolio.mean
        w_hat = quadprog.solve_qp(Q.numpy(), -q.numpy(), C=self.CCmat, b=self.bvec, meq=self.meq)[0]
        #w_hat = qp.solve(Qmat=Q.numpy(), qvec=q.numpy(), Cmat=self.Cmat, cvec=self.cvec,
        #                 Dmat=self.Dmat, dvec=self.dvec, w0=wk, maxiter=self.maxiter,
        #                 tol=self.wtol)
        self.portfolio.weights = wk + self.gamma * (w_hat - wk)
        fun_next = self.get_objective_function_value()
        self.objective_function.append(fun_next.numpy())
        has_w_converged = (tf.abs(self.portfolio.weights - wk) <=
                           .5 * self.wtol * (tf.abs(self.portfolio.weights) +
                                             tf.abs(wk))).numpy().all()
        has_fun_converged = (tf.abs(self._funk - fun_next) <=
                             .5 * self.funtol * (tf.abs(self._funk) +
                                                 tf.abs(fun_next))).numpy().all()
        if has_w_converged or has_fun_converged:
            return False
        self.gamma = self.gamma * (1 - self.zeta * self.gamma)
        self._funk = fun_next
        return True

    def solve(self):
        i = 0
        with tqdm(total=self.maxiter) as pbar:
            while(self.iterate() and i < self.maxiter):
                i += 1
                pbar.update()


def project_line_and_box(weights, lower_bound, upper_bound):
    def objective_function(variable, weights):
        pass
    pass



from __future__ import absolute_import
import os
PACKAGEDIR = os.path.abspath(os.path.dirname(__file__))

from .version import __version__
from .rpp import *
from .sca import *
from .riskfunctions import *
from .vanilla import *



import warnings
import tensorflow as tf
import numpy as np
from .sca import SuccessiveConvexOptimizer
from .riskfunctions import RiskContribOverBudgetDoubleIndex, RiskContribOverVarianceMinusBudget


__all__ = ['RiskParityPortfolio']


class RiskParityPortfolioValidator:
    def __init__(self):
        pass

class RiskParityPortfolio:
    """Designs risk parity portfolios by solving the following optimization problem

    minimize R(w) - alpha * mu.T * w + lambda * w.T Sigma w
    subject to Cw = c, Dw <= d

    where R is a risk concentration function, and alpha and beta are trade-off
    parameters for the expected return and the variance, respectively.

    Parameters
    ----------
    covariance : array, shape=(n, n)
        covariance matrix of the assets
    budget : array, shape=(n,)
        risk budget vector
    equality_constraints : string
        the equality constraint expression
    inequality_constraints : string
        the inequality constraint expression
    weights : array, shape=(n,)
        weights of the portfolio
    risk_concentration : class
        any valid child class of RiskConcentrationFunction
    """

    def __init__(self, covariance, budget=None,
                 equality_constraints=None,
                 inequality_constraints=None,
                 weights=None,
                 risk_concentration=None):
        self.covariance = covariance
        self.budget = budget
        self.weights = weights
        self.risk_concentration = risk_concentration
        self.validate()
        self.has_variance = False
        self.has_mean_return = False

    def get_diag_solution(self):
        w = np.sqrt(self.budget.numpy()) / np.sqrt(np.diagonal(self.covariance.numpy()))
        return w / w.sum()

    @property
    def mean_return(self):
        if self.has_mean_return:
            return tf.tensordot(self.weights, self.mean, axes=1)
        else:
            raise ValueError("the portfolio mean has not been specified, please use add_mean_return")

    @property
    def volatility(self):
        return tf.sqrt(tf.tensordot(self.weights,
                                    tf.linalg.matvec(self.covariance,
                                                     self.weights), axes=1))
    @property
    def risk_contributions(self):
        rc = tf.tensordot(self.weights, tf.multiply(self.covariance, self.weights), axes=1)
        return rc / tf.reduce_sum(rc)

    @property
    def weights(self):
        return self._weights

    @weights.setter
    def weights(self, value):
        if value is None:
            self._weights = tf.ones(self.number_of_assets, dtype=tf.float64) / self.number_of_assets
        else:
            try:
                self._weights = tf.convert_to_tensor(value)
            except Exception as e:
                raise e

    @property
    def number_of_assets(self):
        return self.covariance.shape[0]

    @property
    def risk_concentration(self):
        return self._risk_concentration

    @risk_concentration.setter
    def risk_concentration(self, value):
        if value is None:
            self._risk_concentration = RiskContribOverVarianceMinusBudget(self)
        elif isinstance(value, RiskConcetrationFunction):
            self._risk_concentration = value(self)
        else:
            raise ValueError("risk_concentration {} is not a valid child class"
                             "of RiskConcentrationFunction".format(value))

    @property
    def budget(self):
        return self._budget

    @budget.setter
    def budget(self, value):
        if value is None:
            self._budget = tf.ones(self.number_of_assets, dtype=tf.float64) / self.number_of_assets
        else:
            try:
                self._budget = tf.convert_to_tensor(value)
            except Exception as e:
                raise e

    @property
    def covariance(self):
        return self._covariance

    @covariance.setter
    def covariance(self, value):
        if value.shape[0] != value.shape[1]:
            raise ValueError("shape mismatch: covariance matrix is not a square matrix")
        else:
            try:
                self._covariance = tf.convert_to_tensor(value)
            except Exception as e:
                raise e
            eigvals = np.linalg.eigvals(self._covariance.numpy())
            eigvals = np.sort(eigvals)
            if abs(eigvals[0] / eigvals[-1]) < 1e-6:
                warnings.warn("covariance matrix maybe singular")

    @property
    def equality_constraints(self):
        return self._equality_constraints

    @property
    def inequality_constraints(self):
        return self._inequality_constraints

    @property
    def box_constraints(self):
        return self._box_constraints

    def validate(self):
        if self.covariance.shape[0] != self.budget.shape[0]:
            raise ValueError("shape mismatch between covariance matrix and budget vector")

    def add_mean_return(self, alpha, mean):
        """Whether to consider the maximization of the mean return of the portfolio.

        Parameters
        ----------
        alpha : float
            Hyperparameter associated with the mean return
        mean : array, shape=(n,)
            Vector of means
        """
        self.alpha = alpha
        self.mean = mean
        self.has_mean_return = True

    def add_variance(self, lmd):
        """Whether to consider the minimization of the variance of the portfolio.

        Parameters
        ----------
        lmd : float
            Hyperparameter associated with the variance
        """
        self.lmd = lmd
        self.has_variance = True

    def design(self, **kwargs):
        """Optimize the portfolio.

        Parameters
        ----------
        kwargs : dict
            Dictionary of parameters to be passed to SuccessiveConvexOptimizer.
        """
        self.sca = SuccessiveConvexOptimizer(self, **kwargs)
        self.sca.solve()
